#!/usr/bin/env python
# vim: et ai ts=4 sw=4:

# Copyright 2019-2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import random
import subprocess
import time

from multiprocessing import cpu_count

from charmhelpers.core import hookenv
from charmhelpers.core import host
from charmhelpers.core.templating import render
import distro

from ols import base, wsgi, http
from ols.render import render_logrotate


###############################################################################
# Supporting functions
###############################################################################


def sanitize(s):  # pragma: no cover
    s = s.replace(':', '_')
    s = s.replace('-', '_')
    s = s.replace('/', '_')
    s = s.replace('"', '_')
    s = s.replace("'", '_')
    return s


def sanitized_service_name():  # pragma: no cover
    return sanitize(base.service_name())


def process_template(template_name, template_vars, destination):
    # deferred import so install hook can run to install jinja2
    from jinja2 import Environment, FileSystemLoader

    path = os.path.join(os.environ['CHARM_DIR'], 'templates')
    template_env = Environment(loader=FileSystemLoader(path))

    template = template_env.get_template(template_name).render(template_vars)

    with open(destination, 'w') as inject_file:
        inject_file.write(str(template))


def _random_wait(service):
    charm_config = hookenv.config()
    maxwait = int(charm_config['wsgi_restart_maxwait'])
    if maxwait:
        wait_seconds = random.randint(0, maxwait)
        if wait_seconds:
            current_status, current_message = hookenv.status_get()
            hookenv.status_set(
                "maintenance", "Waiting {}s to restart {}".format(wait_seconds, service)
            )
            time.sleep(wait_seconds)
            hookenv.status_set(current_status, current_message)


def reload_or_restart(service):
    _random_wait(service)
    subprocess.check_call(['systemctl', 'reload-or-restart', service])


def enable_service(service):
    # systemd daemon-reload is done as part of enable
    # http://manpages.ubuntu.com/manpages/xenial/en/man1/systemctl.1.html#commands
    subprocess.check_call(['systemctl', 'enable', service])


@host.restart_on_change(
    {
        "/etc/rsyslog.d/22-gunicorn.conf": ["rsyslog"],
        "/lib/systemd/system/gunicorn.service": ["gunicorn"],
    },
    restart_functions={"rsyslog": reload_or_restart, "gunicorn": enable_service},
)
def configure(working_dir, port, user, group, env_extra, access_logfile='-'):
    wsgi_config = wsgi.layer_options()
    # set charm config options into wsgi_config dict
    charm_config = hookenv.config()
    for conf_name in (
        'wsgi_worker_class',
        'wsgi_workers',
        'wsgi_backlog',
        'wsgi_timeout',
        'wsgi_keep_alive',
        'wsgi_max_requests',
        'wsgi_logrotate_count',
        'wsgi_reuse_port',
    ):
        wsgi_config[conf_name.replace('wsgi_', '')] = charm_config[conf_name]

    wsgi_config['working_dir'] = working_dir
    service_name = sanitized_service_name()
    wsgi_config['service_name'] = service_name

    # hardcoded job for now, we might want to use something else in future
    upstart_file = "/etc/init/gunicorn.conf"
    systemd_file = "/lib/systemd/system/gunicorn.service"
    talisker_syslog_file = "/etc/rsyslog.d/22-gunicorn.conf"
    env_file = wsgi.environment_file()
    # can't use /etc/gunicorn.d/ as that is for debians method
    config_file = wsgi.gunicorn_config_file()
    runner_file = "/srv/gunicorn/run_{}.sh".format(service_name)
    wsgi_config['config_file'] = config_file
    wsgi_config['env_file'] = env_file

    wsgi_config['port'] = port
    wsgi_config['user'] = user
    wsgi_config['group'] = group
    wsgi_config['access_logfile'] = access_logfile

    if str(wsgi_config['workers']) == '0':
        wsgi_config['workers'] = (cpu_count() * 2) + 1

    wsgi_config['env_extra'] = sorted(env_extra.items())

    # gunicorn --pythonpath requires comma separated paths
    comma_separated = wsgi_config.get('python_path', '').replace(':', ',')
    wsgi_config['python_path_comma'] = comma_separated

    host.mkdir('/srv/gunicorn', force=True)
    render(source='config.tmpl', target=config_file, context=wsgi_config)
    hookenv.log('written gunicorn config for %s to %s' % (service_name, config_file))
    render(
        source='runner.tmpl',
        target=runner_file,
        context=wsgi_config,
        group=base.user(),
        perms=0o755,
    )
    if distro.linux_distribution()[-1] == 'trusty':
        render(source='upstart.tmpl', target=upstart_file, context=wsgi_config)
        # upstart job sends logs to /var/log/upstart/gunicorn.log, which is
        # automatically rotated
        hookenv.log('written gunicorn upstart config to %s' % upstart_file)
    else:
        log_path = wsgi_config['log_path'] = '/var/log/gunicorn.log'
        render(source='systemd.tmpl', target=systemd_file, context=wsgi_config)
        render(source='wsgi_env_vars.tmpl', target=env_file, context=wsgi_config)
        # render syslog config to get talisker logs on disk
        render(
            source='wsgi_syslog.tmpl', target=talisker_syslog_file, context=wsgi_config
        )
        render_logrotate('/etc/logrotate.d/gunicorn', wsgi_config)
        # systemd decided to revert group write perms on /var/log, so...
        if not os.path.exists(log_path):
            host.write_file(
                log_path, b'', owner='syslog', group=wsgi_config['group'], perms=0o644
            )
        hookenv.log('written gunicorn systemd config %s' % systemd_file)


def restart(soft=False):
    if soft:
        reload_or_restart('gunicorn')
    else:
        _random_wait('gunicorn')
        host.service_restart('gunicorn')
    try:
        base.check_port('ols.{}.wsgi'.format(base.service_name()), http.port())
    except subprocess.CalledProcessError:  # pragma: no cover
        pass  # already open
