# Copyright 2019-2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import yaml

from charms.reactive import remove_state, when, when_not
from charms.reactive.helpers import any_hook

from ols.base import (
    service_name,
    render_template,
    reset_extra_trusted_networks,
    set_extra_trusted_networks,
)
from ols.http import (
    port,
    metrics_path,
    scrape_interval,
    scrape_timeout,
    active_flag_enabled,
)

from charmhelpers.core import hookenv

# states #


def template_config(services_template, context):
    context['unit_private_address'] = hookenv.unit_private_ip()
    context['port'] = port()
    context['local_unit'] = hookenv.local_unit()
    context['service_name'] = service_name()
    return render_template(services_template, context)


@when('loadbalancer.available')
def configure_loadbalancer(lb):
    # XXX: set workload status?
    config = hookenv.config()
    haproxy_template = config.get('haproxy_services')

    if haproxy_template:
        # legacy templating
        services_yaml = template_config(haproxy_template, config.copy())

        try:
            yaml.safe_load(services_yaml)
        except Exception:
            hookenv.log('Could not parse templated haproxy yaml')
            hookenv.status_set('blocked', 'bad haproxy_services yaml template config')
            return

    else:
        try:
            options = yaml.safe_load(config.get('haproxy_service_options'))
        except Exception:
            hookenv.log('could not parse haproxy_service_options yaml')
            hookenv.status_set('blocked', 'bad haproxy_service_options yaml config')
            return

        server_options = config.get('haproxy_server_options')
        server = [
            hookenv.local_unit().replace('/', '-'),
            hookenv.unit_private_ip(),
            port(),
        ]

        # expose a specific backend for targetting this specific
        # serveice. This is not routed by default, but can be access via
        # headers if the haproxy config is setup appropriately.
        backend_server = list(server)
        backend_server.append(server_options)

        # handle optional active flag, weighting the servers if needed.
        if active_flag_enabled():
            weight = '1' if config.get('active') else '0'
            server.append("{} weight {}".format(server_options, weight))
        else:
            server.append(server_options)

        services = [
            {
                'service_name': service_name(),
                'service_port': port(),
                'service_host': '0.0.0.0',
                'service_options': options,
                'servers': [server.copy()],  # use copy to avoid yaml references
            }
        ]

        juju_svc = hookenv.service_name()

        # if we are a switch service, add addtional backend for targetting this
        # specific service
        if juju_svc != service_name():
            # note: this will go into the default haproxy frontend for this
            # service, which is not usually used, as we create a manual TLS
            # frontend in the haproxy charm config.
            services[0]['service_options'].append(
                'use_backend {0} if {{ hdr(x-use-backend) -i {0} }}'.format(juju_svc)
            )

            services[0]['backends'] = [
                {'backend_name': juju_svc, 'servers': [backend_server.copy()]}
            ]

        services_yaml = yaml.safe_dump(services)

    for rel in hookenv.relations_of_type('loadbalancer'):
        hookenv.relation_set(rel['__relid__'], services=services_yaml)


@when('nrpe-external-master.available')
def configure_nrpe(nrpe):
    hookenv.log("Updating nrpe configs")
    config = hookenv.config()
    for rel_id in hookenv.relation_ids("nrpe-external-master"):
        hookenv.relation_set(rel_id, primary=True)
    context = config['nagios_context']
    nrpe.add_check(
        [
            "/usr/lib/nagios/plugins/check_http",
            "-I",
            "127.0.0.1",
            "-p",
            str(port()),
            "-e",
            "' 200 OK'",
            "-H",
            "localhost",
            "-u",
            "'/_status/check'",
        ],
        context=context,
        servicegroups=config.get('nagios_servicegroups', context),
        name="{}-{}".format(service_name(), config['environment']),
        description="{} server {}".format(service_name(), config['environment']),
        unit=hookenv.local_unit(),
    )
    # XXX: set workload status?


@when('prometheus.available')
def configure_prometheus(prometheus):
    if any_hook('{provides:prometheus}-relation-{joined,changed}'):
        params = {
            'port': port(),
            'path': metrics_path(),
            'scrape_interval': scrape_interval(),
            'scrape_timeout': scrape_timeout(),
        }
        prometheus.configure(**params)

        set_extra_trusted_networks(
            'prometheus', [prometheus.get_remote('private-address')]
        )
        remove_state('service.configured')  # Force service reconfiguration.


@when_not('prometheus.available')
def unconfigure_prometheus(*args, **kwargs):
    if any_hook('{provides:prometheus}-relation-departed'):
        reset_extra_trusted_networks('prometheus')
        remove_state('service.configured')  # Force service reconfiguration.
