# Copyright 2019-2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import urllib

import yaml

import reactive.ols_http

from unittest.mock import patch, Mock, call

from charms import layer
from charms.reactive import bus, set_flag, clear_flag

from ols import http

from base_fixtures import CharmEnvFixture, ReactiveCharmEnvFixture, BaseTestCase


class HttpTest(BaseTestCase):
    def setUp(self):
        super().setUp()
        self.charm_dir = '/charm'
        self.fakes = self.useFixture(CharmEnvFixture(self.charm_dir))

    def test_layer_options(self):
        self.assertEqual(http._layer_options(), layer.options().get('ols-http', {}))

    def test_port(self):
        self.assertEqual(http.port(), '8000')

    def test_metrics_path(self):
        self.assertEqual(http.metrics_path(), '/_status/metrics')

    def test_scrape_interval(self):
        self.assertEqual(http.scrape_interval(), '1m')

    def test_scrape_timeout(self):
        self.assertEqual(http.scrape_timeout(), '10s')

    def test_active_flag_enabled(self):
        self.assertEqual(http.active_flag_enabled(), False)

    @patch.object(urllib.request, 'urlopen')
    def test_is_listening(self, urlopen):
        urlopen.return_value = True
        self.assertTrue(http.is_listening(http_port=8080))


class ReactiveTest(BaseTestCase):

    maxDiff = 2000

    def setUp(self):
        super().setUp()
        self.charm_dir = '/charm'
        self.fakes = self.useFixture(ReactiveCharmEnvFixture(self.charm_dir))

    def yaml_relation_data(self, mock_set, id, key):
        """Load the yaml string for key in relation data."""
        for args, kwargs in mock_set.call_args_list:
            if args[0] == id:
                return yaml.safe_load(kwargs[key])
        raise AssertionError('relation id {} not found in mock calls'.format(id))

    @patch.object(reactive.ols_http, 'render_template')
    @patch.object(reactive.ols_http.hookenv, 'unit_private_ip')
    @patch.object(reactive.ols_http.hookenv, 'local_unit')
    def test_template_config(self, local_unit, unit_private_ip, render_template):
        unit_private_ip.return_value = '10.10.0.1'
        local_unit.return_value = 'testservice/0'
        reactive.ols_http.template_config("", {})
        expected_context = {
            'unit_private_address': "10.10.0.1",
            'port': '8000',
            'local_unit': 'testservice/0',
            'service_name': 'testservice',
        }
        render_template.assert_called_once_with("", expected_context)

    @patch.object(reactive.ols_http.hookenv, 'relation_set')
    @patch.object(reactive.ols_http.hookenv, 'relations_of_type')
    @patch.object(reactive.ols_http.hookenv, 'service_name')
    @patch.object(reactive.ols_http.hookenv, 'unit_private_ip')
    @patch.object(reactive.ols_http.hookenv, 'local_unit')
    def test_configure_loadbalancer(
        self, local_unit, unit_private_ip, service_name, rels_of_type, rel_set
    ):
        unit_private_ip.return_value = '10.10.0.1'
        local_unit.return_value = 'testservice/0'
        service_name.return_value = 'testservice'
        rels_of_type.return_value = [{'__relid__': 'relid'}]
        interface = Mock()
        set_flag('loadbalancer.available')
        reactive.ols_http.configure_loadbalancer(interface)
        haproxy_service_config = {
            'service_name': 'testservice',
            'servers': [
                [
                    'testservice-0',
                    '10.10.0.1',
                    '8000',
                    'check inter 3000 rise 2 fall 3 maxconn 100',
                ]
            ],
            'service_options': [
                'mode http',
                'option httplog',
                'option httpchk HEAD /_status/check',
                'balance roundrobin',
                'timeout client 5000',
                'timeout server 5000',
                'capture request header X-Request-Id len 36',
                'option forwardfor',
            ],
            'service_host': '0.0.0.0',
            'service_port': '8000',
        }
        rel_data = self.yaml_relation_data(rel_set, 'relid', 'services')[0]
        self.assertDictEqual(rel_data, haproxy_service_config)

    @patch.object(reactive.ols_http.hookenv, 'relation_set')
    @patch.object(reactive.ols_http.hookenv, 'relations_of_type')
    @patch.object(reactive.ols_http.hookenv, 'service_name')
    @patch.object(reactive.ols_http.hookenv, 'unit_private_ip')
    @patch.object(reactive.ols_http.hookenv, 'local_unit')
    def test_configure_loadbalancer_different_juju_svc_name(
        self, local_unit, unit_private_ip, service_name, rels_of_type, rel_set
    ):
        unit_private_ip.return_value = '10.10.0.1'
        local_unit.return_value = 'testservice/0'
        service_name.return_value = 'adifferentnamefromthecharm'
        rels_of_type.return_value = [{'__relid__': 'relid'}]
        interface = Mock()
        set_flag('loadbalancer.available')
        reactive.ols_http.configure_loadbalancer(interface)
        haproxy_service_config = {
            'backends': [
                {
                    'backend_name': 'adifferentnamefromthecharm',
                    'servers': [
                        [
                            'testservice-0',
                            '10.10.0.1',
                            '8000',
                            'check inter 3000 rise 2 fall 3 maxconn 100',
                        ]
                    ],
                }
            ],
            'service_name': 'testservice',
            'servers': [
                [
                    'testservice-0',
                    '10.10.0.1',
                    '8000',
                    'check inter 3000 rise 2 fall 3 maxconn 100',
                ]
            ],
            'service_options': [
                'mode http',
                'option httplog',
                'option httpchk HEAD /_status/check',
                'balance roundrobin',
                'timeout client 5000',
                'timeout server 5000',
                'capture request header X-Request-Id len 36',
                'option forwardfor',
                'use_backend adifferentnamefromthecharm if '
                '{ hdr(x-use-backend) -i adifferentnamefromthecharm }',
            ],
            'service_host': '0.0.0.0',
            'service_port': '8000',
        }
        rel_data = self.yaml_relation_data(rel_set, 'relid', 'services')[0]
        self.assertDictEqual(rel_data, haproxy_service_config)

    @patch.object(reactive.ols_http, 'active_flag_enabled')
    @patch.object(reactive.ols_http.hookenv, 'relation_set')
    @patch.object(reactive.ols_http.hookenv, 'relations_of_type')
    @patch.object(reactive.ols_http.hookenv, 'service_name')
    @patch.object(reactive.ols_http.hookenv, 'unit_private_ip')
    @patch.object(reactive.ols_http.hookenv, 'local_unit')
    def test_configure_loadbalancer_active_flag(
        self,
        local_unit,
        unit_private_ip,
        service_name,
        rels_of_type,
        rel_set,
        active_enabled,
    ):
        unit_private_ip.return_value = '10.10.0.1'
        local_unit.return_value = 'testservice/0'
        service_name.return_value = 'testservice'
        rels_of_type.return_value = [{'__relid__': 'relid'}]
        active_enabled.return_value = True
        self.fakes.config.update({'active': True})
        interface = Mock()
        set_flag('loadbalancer.available')
        reactive.ols_http.configure_loadbalancer(interface)
        haproxy_service_config = {
            'service_name': 'testservice',
            'servers': [
                [
                    'testservice-0',
                    '10.10.0.1',
                    '8000',
                    'check inter 3000 rise 2 fall 3 maxconn 100 weight 1',
                ]
            ],
            'service_options': [
                'mode http',
                'option httplog',
                'option httpchk HEAD /_status/check',
                'balance roundrobin',
                'timeout client 5000',
                'timeout server 5000',
                'capture request header X-Request-Id len 36',
                'option forwardfor',
            ],
            'service_host': '0.0.0.0',
            'service_port': '8000',
        }
        rel_data = self.yaml_relation_data(rel_set, 'relid', 'services')[0]
        self.assertDictEqual(rel_data, haproxy_service_config)

    @patch.object(reactive.ols_http, 'active_flag_enabled')
    @patch.object(reactive.ols_http.hookenv, 'relation_set')
    @patch.object(reactive.ols_http.hookenv, 'relations_of_type')
    @patch.object(reactive.ols_http.hookenv, 'service_name')
    @patch.object(reactive.ols_http.hookenv, 'unit_private_ip')
    @patch.object(reactive.ols_http.hookenv, 'local_unit')
    def test_configure_loadbalancer_active_flag_and_different_name(
        self,
        local_unit,
        unit_private_ip,
        service_name,
        rels_of_type,
        rel_set,
        active_enabled,
    ):
        unit_private_ip.return_value = '10.10.0.1'
        local_unit.return_value = 'testservice/0'
        service_name.return_value = 'adifferentnamefromthecharm'
        rels_of_type.return_value = [{'__relid__': 'relid'}]
        active_enabled.return_value = True
        self.fakes.config.update({'active': True})
        interface = Mock()
        set_flag('loadbalancer.available')
        reactive.ols_http.configure_loadbalancer(interface)
        haproxy_service_config = {
            'backends': [
                {
                    'backend_name': 'adifferentnamefromthecharm',
                    'servers': [
                        [
                            'testservice-0',
                            '10.10.0.1',
                            '8000',
                            'check inter 3000 rise 2 fall 3 maxconn 100',
                        ]
                    ],
                }
            ],
            'service_name': 'testservice',
            'servers': [
                [
                    'testservice-0',
                    '10.10.0.1',
                    '8000',
                    'check inter 3000 rise 2 fall 3 maxconn 100 weight 1',
                ]
            ],
            'service_options': [
                'mode http',
                'option httplog',
                'option httpchk HEAD /_status/check',
                'balance roundrobin',
                'timeout client 5000',
                'timeout server 5000',
                'capture request header X-Request-Id len 36',
                'option forwardfor',
                'use_backend adifferentnamefromthecharm if '
                '{ hdr(x-use-backend) -i adifferentnamefromthecharm }',
            ],
            'service_host': '0.0.0.0',
            'service_port': '8000',
        }
        rel_data = self.yaml_relation_data(rel_set, 'relid', 'services')[0]
        self.assertDictEqual(rel_data, haproxy_service_config)

    @patch.object(reactive.ols_http.hookenv, 'service_name')
    @patch.object(reactive.ols_http.hookenv, 'unit_private_ip')
    @patch.object(reactive.ols_http.hookenv, 'local_unit')
    def test_configure_loadbalancer_syntax_error(
        self, local_unit, unit_private_ip, service_name
    ):
        interface = Mock()
        set_flag('loadbalancer.available')
        haproxy_services = '{this is a test with invalid yaml'
        self.fakes.config.update({'haproxy_service_options': haproxy_services})
        reactive.ols_http.configure_loadbalancer(interface)
        self.fakes.log.assert_has_calls(
            [call('could not parse haproxy_service_options yaml')]
        )
        self.fakes.status_set.assert_called_once_with(
            'blocked', 'bad haproxy_service_options yaml config'
        )

    @patch.object(reactive.ols_http.hookenv, 'relation_set')
    @patch.object(reactive.ols_http.hookenv, 'relations_of_type')
    @patch.object(reactive.ols_http.hookenv, 'service_name')
    @patch.object(reactive.ols_http.hookenv, 'unit_private_ip')
    @patch.object(reactive.ols_http.hookenv, 'local_unit')
    def test_configure_loadbalancer_legacy_templating(
        self, local_unit, unit_private_ip, service_name, rels_of_type, rel_set
    ):
        unit_private_ip.return_value = '10.10.0.1'
        local_unit.return_value = 'testservice/0'
        service_name.return_value = 'testservice'
        rels_of_type.return_value = [{'__relid__': 'relid'}]
        interface = Mock()
        set_flag('loadbalancer.available')
        haproxy_services = 'this is a test'
        self.fakes.config.update({'haproxy_services': haproxy_services})
        reactive.ols_http.configure_loadbalancer(interface)
        rel_set.assert_called_once_with('relid', services=haproxy_services)

    @patch.object(reactive.ols_http.hookenv, 'service_name')
    @patch.object(reactive.ols_http.hookenv, 'unit_private_ip')
    @patch.object(reactive.ols_http.hookenv, 'local_unit')
    def test_configure_loadbalancer_legacy_templating_syntax_error(
        self, local_unit, unit_private_ip, service_name
    ):
        interface = Mock()
        set_flag('loadbalancer.available')
        haproxy_services = '{this is a test with invalid yaml'
        self.fakes.config.update({'haproxy_services': haproxy_services})
        reactive.ols_http.configure_loadbalancer(interface)
        self.fakes.log.assert_has_calls(
            [call('Could not parse templated haproxy yaml')]
        )
        self.fakes.status_set.assert_called_once_with(
            'blocked', 'bad haproxy_services yaml template config'
        )

    @patch.object(reactive.ols_http.hookenv, 'unit_private_ip')
    @patch.object(reactive.ols_http.hookenv, 'local_unit')
    @patch.object(reactive.ols_http.hookenv, 'relation_set')
    @patch.object(reactive.ols_http.hookenv, 'relation_ids')
    def test_configure_nrpe(self, rel_ids, rel_set, local_unit, unit_private_ip):
        unit_private_ip.return_value = '10.10.0.1'
        local_unit.return_value = 'testservice/0'
        rel_ids.return_value = ['rel-id']
        interface = Mock()
        add_check = Mock()
        interface.add_check = add_check
        set_flag('nrpe-external-master.available')
        reactive.ols_http.configure_nrpe(interface)
        rel_ids.assert_called_once_with('nrpe-external-master')
        rel_set.assert_called_once_with('rel-id', primary=True)
        add_check.assert_called_once_with(
            [
                '/usr/lib/nagios/plugins/check_http',
                '-I',
                '127.0.0.1',
                '-p',
                '8000',
                '-e',
                "' 200 OK'",
                '-H',
                'localhost',
                '-u',
                "'/_status/check'",
            ],
            context='',
            description='testservice server production',
            name='testservice-production',
            servicegroups='',
            unit='testservice/0',
        )

    @patch.object(reactive.ols_http, 'set_extra_trusted_networks')
    @patch.object(reactive.ols_http, 'any_hook')
    def test_configure_prometheus(self, any_hook, set_extra_trusted_networks):
        any_hook.return_value = True
        interface = Mock()
        interface.configure = Mock()
        interface.get_remote = Mock()
        interface.get_remote.return_value = '10.10.0.1'
        set_flag('prometheus.available')
        set_flag('service.configured')
        reactive.ols_http.configure_prometheus(interface)
        params = {
            'port': http.port(),
            'path': http.metrics_path(),
            'scrape_interval': http.scrape_interval(),
            'scrape_timeout': http.scrape_timeout(),
        }
        interface.configure.assert_called_once_with(**params)
        self.assertNotIn('service.configured', bus.get_states().keys())
        set_extra_trusted_networks.assert_called_once_with('prometheus', ['10.10.0.1'])

    @patch.object(reactive.ols_http, 'reset_extra_trusted_networks')
    @patch.object(reactive.ols_http, 'any_hook')
    def test_unconfigure_prometheus(self, any_hook, reset_extra_trusted_networks):
        any_hook.return_value = True
        set_flag('service.configured')
        clear_flag('prometheus.available')
        reactive.ols_http.unconfigure_prometheus()
        reset_extra_trusted_networks.called_once_with('prometheus')
        self.assertNotIn('service.configured', bus.get_states().keys())
