# Copyright 2019-2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest.mock import patch, call

from charms import layer

from ols import base, wsgi, gunicorn

from base_fixtures import CharmEnvFixture, ReactiveCharmEnvFixture, BaseTestCase


class WsgiTest(BaseTestCase):
    def setUp(self):
        super().setUp()
        self.charm_dir = "/charm"
        self.fakes = self.useFixture(CharmEnvFixture(self.charm_dir))

    def test_layer_options(self):
        self.assertEqual(wsgi.layer_options(), layer.options().get("ols-wsgi", {}))

    def test_status_path(self):
        self.assertEqual(wsgi.status_path(), "")

    def test_gunicorn_config_file(self):
        self.assertEqual(
            wsgi.gunicorn_config_file(),
            "/srv/gunicorn/{}.py".format(base.service_name()),
        )

    def test_environment_file(self):
        self.assertEqual(wsgi.environment_file(), "/srv/gunicorn/environment_variables")

    def test_config_files(self):
        expected = [
            '/srv/testservice/etc/service.conf',
            '/etc/init/gunicorn.conf',
            '/srv/gunicorn/testservice.py',
            '/srv/gunicorn/environment_variables',
        ]
        self.assertEqual(wsgi.config_files(), expected)


class GunicornTest(BaseTestCase):
    def setUp(self):
        super().setUp()
        self.charm_dir = "/charm"
        self.fakes = self.useFixture(ReactiveCharmEnvFixture(self.charm_dir))

    @patch.object(gunicorn, "_random_wait")
    @patch.object(gunicorn.subprocess, "check_call")
    def test_reload_or_restart(self, check_call, random_wait):
        service = "foobar"
        gunicorn.reload_or_restart(service)

        random_wait.assert_called_once()
        check_call.assert_called_once_with(["systemctl", "reload-or-restart", service])

    @patch.object(gunicorn.base, "check_port")
    @patch.object(gunicorn, "_random_wait")
    @patch.object(gunicorn.subprocess, "check_call")
    def test_restart_soft(self, check_call, random_wait, check_port):
        gunicorn.restart(soft=True)

        random_wait.assert_called_once()
        check_call.assert_called_once_with(
            ["systemctl", "reload-or-restart", "gunicorn"]
        )
        check_port.assert_called_once_with("ols.testservice.wsgi", '8000')

    @patch.object(gunicorn.hookenv, "status_get")
    @patch.object(gunicorn.hookenv, "status_set")
    @patch.object(gunicorn.random, "randint")
    @patch.object(gunicorn.time, "sleep")
    @patch.object(gunicorn.subprocess, "check_call")
    def test__random_wait(self, check_call, sleep, randint, status_set, status_get):
        service = "foobar"
        new_config = {"wsgi_restart_maxwait": 50}
        self.fakes.config.update(new_config)
        randint.return_value = 42
        status_get.return_value = ("active", "ready")

        gunicorn.reload_or_restart(service)

        randint.assert_called_once()
        status_get.assert_called_once()
        status_set.assert_has_calls(
            [
                call("maintenance", "Waiting 42s to restart foobar"),
                call("active", "ready"),
            ]
        )
        sleep.assert_called_once_with(42)
        check_call.assert_called_once_with(["systemctl", "reload-or-restart", service])

    @patch.object(gunicorn.hookenv, "status_get")
    @patch.object(gunicorn.hookenv, "status_set")
    @patch.object(gunicorn.random, "randint")
    @patch.object(gunicorn.time, "sleep")
    @patch.object(gunicorn.subprocess, "check_call")
    def test__random_wait_maxwait_0(
        self, check_call, sleep, randint, status_set, status_get
    ):
        service = "foobar"
        new_config = {"wsgi_restart_maxwait": 0}
        self.fakes.config.update(new_config)

        gunicorn.reload_or_restart(service)

        randint.assert_not_called()
        status_get.assert_not_called()
        status_set.assert_not_called()
        sleep.assert_not_called()
        check_call.assert_called_once_with(["systemctl", "reload-or-restart", service])
